#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""The module provides support for Root Key Hash table."""

import logging
import math
from typing import List, Optional, Union

from spsdk.crypto import (
    Certificate,
    EllipticCurvePublicKey,
    PrivateKey,
    PublicKey,
    RSAPublicKey,
    _PrivateKeyTuple,
    _PublicKeyTuple,
)
from spsdk.crypto.loaders import extract_public_key, extract_public_key_from_data
from spsdk.exceptions import SPSDKError
from spsdk.utils.crypto.common import crypto_backend

logger = logging.getLogger(__name__)


class RKHT:
    """Root Key Hash Table class."""

    def __init__(
        self,
        keys: Optional[List] = None,
        keys_cnt: int = 4,
        min_keys_cnt: int = 4,
        password: Optional[str] = None,
        search_paths: Optional[List[str]] = None,
    ) -> None:
        """Initialization of Root Key Hash Table class.

        :param keys: List of source of root keys (The public keys could get also from private key
            or certificates), defaults to None
        :param password: Optional password to open secured private keys, defaults to None.
        :param search_paths: List of paths where to search for the file, defaults to None
        """
        self.keys_cnt = keys_cnt
        self.min_keys_cnt = min_keys_cnt
        self.rotk = (
            [RKHT.convert_key(x, password, search_paths=search_paths) for x in keys] if keys else []
        )

    @staticmethod
    def _hash_algorithm_output_size(key: PublicKey) -> int:
        """Get Hash algorithm output size for the key.

        :param key: Key to get hash.
        :raises SPSDKError: Invalid kye type.
        :return: Size in bits of hash.
        """
        if isinstance(key, EllipticCurvePublicKey):
            return key.key_size

        if isinstance(key, RSAPublicKey):
            # In case of RSA keys, hash is always SHA-256, regardless of the key length
            return 256

        raise SPSDKError("RKHT: Unsupported key type to load.")

    @staticmethod
    def _hash_algorithm(key: PublicKey) -> str:
        """Get Hash algorithm name for the key.

        :param key: Key to get hash.
        :raises SPSDKError: Invalid kye type.
        :return: Name of hash algorithm.
        """
        return f"sha{RKHT._hash_algorithm_output_size(key)}"

    @property
    def hash_algorithm(self) -> str:
        """Used HASH algorithm name."""
        assert len(self.rotk) > 0
        return RKHT._hash_algorithm(self.rotk[0])

    @property
    def hash_algorithm_size(self) -> int:
        """Used HASH algorithm size in bytes."""
        assert len(self.rotk) > 0
        return RKHT._hash_algorithm_output_size(self.rotk[0])

    def validate(self) -> None:
        """Validate the RKHT object."""
        if len(self.rotk) == 0:
            raise SPSDKError("RKHT is missing input public keys.")

        if self.keys_cnt >= len(self.rotk) < self.min_keys_cnt:
            raise SPSDKError(f"RKHT: Invalid key count: ({len(self.rotk)}).")

        if not all(isinstance(x, type(self.rotk[0])) for x in self.rotk):
            raise SPSDKError("RKHT must contains all keys same instances.")
        if not all(RKHT._hash_algorithm(x) == self.hash_algorithm for x in self.rotk):
            raise SPSDKError("RKHT must have same hash algorithm for all keys.")

    @staticmethod
    def calc_key_hash(
        public_key: PublicKey,
        sha_width: int = 256,
    ) -> bytes:
        """Calculate a hash out of public key's exponent and modulus in RSA case, X/Y in EC.

        :param public_key: List of public keys to compute hash from.
        :param sha_width: Used hash algorithm.
        :raises SPSDKError: Unsupported public key type
        :return: Computed hash.
        """
        if isinstance(public_key, RSAPublicKey):
            n_1: int = public_key.public_numbers().e  # type: ignore # MyPy is unable to pickup the class member
            n1_len = math.ceil(n_1.bit_length() / 8)
            n_2: int = public_key.public_numbers().n  # type: ignore # MyPy is unable to pickup the class member
            n2_len = math.ceil(n_2.bit_length() / 8)
        elif isinstance(public_key, EllipticCurvePublicKey):
            n_1: int = public_key.public_numbers().y  # type: ignore # MyPy is unable to pickup the class member
            n1_len = sha_width // 8
            n_2: int = public_key.public_numbers().x  # type: ignore # MyPy is unable to pickup the class member
            n2_len = sha_width // 8
        else:
            raise SPSDKError(f"Unsupported key type: {type(public_key)}")

        n1_bytes = n_1.to_bytes(n1_len, "big")
        n2_bytes = n_2.to_bytes(n2_len, "big")

        return crypto_backend().hash(n2_bytes + n1_bytes, algorithm=f"sha{sha_width}")

    def key_hashes(self) -> List[bytes]:
        """List of individual key hashes.

        :return: List of individual key hashes.
        """
        ret = []
        for i in range(self.keys_cnt):
            if i < len(self.rotk) and self.rotk[i]:
                ret.append(
                    RKHT.calc_key_hash(self.rotk[i], RKHT._hash_algorithm_output_size(self.rotk[i]))
                )
            else:
                ret.append(bytes(RKHT._hash_algorithm_output_size(self.rotk[0]) // 8))
        return ret

    def rotkh(self) -> bytes:
        """Root of Key Table hash.

        :return: Hash of Hashes of public key.
        """
        rotkh = crypto_backend().hash(bytearray().join(self.key_hashes()), self.hash_algorithm)
        logger.info(f"ROTKH: {rotkh.hex()}")
        return rotkh

    def add_key(self, key: PublicKey) -> None:
        """Add additional public key.

        :param key: Root of Trust public key.
        """
        self.rotk.append(key)

    @staticmethod
    def convert_key(
        key: Union[str, bytes, bytearray, PublicKey, PrivateKey, Certificate],
        password: Optional[str] = None,
        search_paths: Optional[List[str]] = None,
    ) -> PublicKey:
        """Convert practically whole input that could hold Public key into public key.

        :param key: Public key in Certificate/Private key, Public key as a path to file,
            loaded bytes or supported class.
        :param password: Optional password to open secured private keys, defaults to None.
        :param search_paths: List of paths where to search for the file, defaults to None
        :raises SPSDKError: Invalid kye type.
        :return: Public Key object.
        """
        if isinstance(key, _PublicKeyTuple):
            return key

        if isinstance(key, _PrivateKeyTuple):
            return key.public_key()

        if isinstance(key, Certificate):
            public_key = key.public_key()
            assert isinstance(public_key, _PublicKeyTuple)
            return public_key

        if isinstance(key, str):
            return extract_public_key(key, password, search_paths=search_paths)

        if isinstance(key, (bytes, bytearray)):
            return extract_public_key_from_data(key, password)

        raise SPSDKError("RKHT: Unsupported key to load.")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Common utils for DAT module."""
import math
from typing import Optional, Union

from spsdk import crypto
from spsdk.crypto import utils_cryptography


def reconstruct_signature(signature_bytes: bytes, size: Optional[int] = None) -> bytes:
    """Reconstructs signature.

    :param signature_bytes: signature's bytes
    :param size: size of r and s bytes (from signature)
    :return: reconstructed signature
    """
    size = len(signature_bytes) // 2
    r_bytes = signature_bytes[:size]
    r = int.from_bytes(r_bytes, "big")
    s_bytes = signature_bytes[size : (size * 2)]
    s = int.from_bytes(s_bytes, "big")
    signature = utils_cryptography.encode_dss_signature(r, s)
    return signature


def ecc_public_numbers_to_bytes(
    public_numbers: crypto.EllipticCurvePublicNumbers, length: Optional[int] = None
) -> bytes:
    """Converts public numbers from ECC key into bytes.

    :param public_numbers: instance of ecc public numbers
    :param length: length of bytes object to use
    :return: bytes representation
    """
    x = public_numbers.x
    y = public_numbers.y
    length = length or math.ceil(x.bit_length() / 8)
    x_bytes = x.to_bytes(length, "big")
    y_bytes = y.to_bytes(length, "big")
    return x_bytes + y_bytes


def rsa_key_to_bytes(
    key: Union[crypto.RSAPublicKey, crypto.RSAPrivateKeyWithSerialization],
    exp_length: Optional[int] = None,
    modulus_length: Optional[int] = None,
) -> bytes:
    """Converts RSA key into bytes.

    :param key: Union of types: RSAPublicKey, RSAPrivateKeyWithSerialization
    :param exp_length: Length of exponent's bytes to use if none it will be calculated
    :param modulus_length: Length of modulus's bytes to use if none it will be calculated
    :return: Combined modulus and exponent bytes
    """
    exp_rotk = key.public_numbers().e  # type: ignore
    mod_rotk = key.public_numbers().n  # type: ignore
    exp_length = exp_length or math.ceil(exp_rotk.bit_length() / 8)
    modulus_length = modulus_length or math.ceil(mod_rotk.bit_length() / 8)
    exp_rotk_bytes = exp_rotk.to_bytes(exp_length, "big")
    mod_rotk_bytes = mod_rotk.to_bytes(modulus_length, "big")
    return mod_rotk_bytes + exp_rotk_bytes


def ecc_key_to_bytes(key: crypto.EllipticCurvePublicKey, length: Optional[int] = None) -> bytes:
    """Converts key into bytes.

    :param key: instance of ECC Public Key
    :param length: length of bytes object to use
    :return: bytes representation
    """
    public_numbers = key.public_numbers()
    return ecc_public_numbers_to_bytes(public_numbers=public_numbers, length=length)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Trust Provisioning HOST application support."""

import logging
from typing import Callable, Optional

from spsdk.utils.misc import Timeout

from . import TpDevInterface

logger = logging.getLogger(__name__)


class TrustProvisioningConfig:
    """Trust provisioning support in none trusted environment."""

    def __init__(
        self,
        tpdev: TpDevInterface,
        info_print: Callable[[str], None],
    ) -> None:
        """Trust Provisioning Host support class.

        :param tpdev: [description]
        :param info_print: [description]
        """
        self.tpdev = tpdev
        self.info_print = info_print

    def upload(
        self, user_config: dict, user_config_dir: Optional[str] = None, timeout: int = 60
    ) -> None:
        """Upload the user data into the provisioning device."""
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP device")
            self.tpdev.open()

            self.info_print("Upload provisioning data to TP target")
            self.tpdev.upload(config_data=user_config, config_dir=user_config_dir)

            self.info_print(
                f"TP device personalization ended correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(
                f"TP device personalization FAILED in {loc_timeout.get_consumed_time_ms()} ms."
            )
            raise
        finally:
            logger.debug("Closing TP device")
            self.tpdev.close()

    def setup(self, timeout: int = 60) -> None:
        """Setup the provisioning device."""
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP device")
            self.tpdev.open()

            self.info_print("Setting up the TP device")
            self.tpdev.setup()
            self.info_print(
                f"TP device setup ended correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(f"TP device setup FAILED in {loc_timeout.get_consumed_time_ms()} ms.")
            raise
        finally:
            logger.debug("Closing TP device")
            self.tpdev.close()

    def seal(self, timeout: int = 60) -> None:
        """Seal the provisioning device."""
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP device")
            self.tpdev.open()

            self.info_print("Sealing the smart card")

            self.tpdev.seal()
            self.info_print(
                f"TP device sealing ended correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(f"TP device sealing FAILED in {loc_timeout.get_consumed_time_ms()} ms.")
            raise
        finally:
            logger.debug("Closing TP device")
            self.tpdev.close()

    def get_counters(self, timeout: int = 60) -> None:
        """Seal the provisioning device."""
        try:
            loc_timeout = Timeout(timeout=timeout, units="s")
            logger.debug("Opening TP device")
            self.tpdev.open()

            self.info_print("Retrieving counters from TP Device.")

            self.info_print(f"Current provisioning counter: {self.tpdev.get_prov_counter():,}")
            self.info_print(f"Provisioning attempts left  : {self.tpdev.get_prov_remainder():,}")
            self.info_print(
                f"TP device counters retrieval ended correctly in {loc_timeout.get_consumed_time_ms()} ms."
            )
        except:
            self.info_print(
                f"TP device counters retrieval FAILED in {loc_timeout.get_consumed_time_ms()} ms."
            )
            raise
        finally:
            logger.debug("Closing TP device")
            self.tpdev.close()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for USB communication with a target using SDP protocol."""

import logging
from typing import List, Optional, Tuple, Union

import libusbsio

from spsdk.utils.misc import get_hash
from spsdk.utils.usbfilter import NXPUSBDeviceFilter, USBDeviceFilter

from ..commands import CmdPacket, CmdResponse
from ..exceptions import SdpConnectionError, SdpError
from .base import SDPInterface

logger = logging.getLogger(__name__)

# import os
# os.environ['PYUSB_DEBUG'] = 'debug'
# os.environ['PYUSB_LOG_FILENAME'] = 'usb.log'

HID_REPORT = {
    # name | id | length
    "CMD": (0x01, 1024, False),
    "DATA": (0x02, 1024, False),
    "HAB": (0x03, 4),
    "RET": (0x04, 64),
}

USB_DEVICES = {
    # NAME    | VID   | PID
    "MX6DQP": (0x15A2, 0x0054),
    "MX6SDL": (0x15A2, 0x0061),
    "MX6SL": (0x15A2, 0x0063),
    "MX6SX": (0x15A2, 0x0071),
    "MX6UL": (0x15A2, 0x007D),
    "MX6ULL": (0x15A2, 0x0080),
    "MX6SLL": (0x15A2, 0x0128),
    "MX7SD": (0x15A2, 0x0076),
    "MX7ULP": (0x1FC9, 0x0126),
    "VYBRID": (0x15A2, 0x006A),
    "MXRT20": (0x1FC9, 0x0130),
    "MXRT50": (0x1FC9, 0x0130),
    "MXRT60": (0x1FC9, 0x0135),
    "MX8MQ": (0x1FC9, 0x012B),
    "MX8QXP-A0": (0x1FC9, 0x007D),
    "MX8QM-A0": (0x1FC9, 0x0129),
    "MX8QXP": (0x1FC9, 0x012F),
    "MX8QM": (0x1FC9, 0x0129),
    "MX815": (0x1FC9, 0x013E),
    "MX865": (0x1FC9, 0x0146),
    "MX93": (0x1FC9, 0x014E),
}


def scan_usb(device_id: Optional[str] = None) -> List["RawHid"]:
    """Scan connected USB devices. Return a list of all devices found.

    :param device_id: see USBDeviceFilter classes constructor for usb_id specification
    :return: list of matching RawHid devices
    """
    usb_filter = NXPUSBDeviceFilter(usb_id=device_id, nxp_device_names=USB_DEVICES)
    return RawHid.enumerate(usb_filter)


########################################################################################################################
# USB HID Interface Class
########################################################################################################################
class RawHid(SDPInterface):
    """Base class for OS specific RAW HID Interface classes."""

    @property
    def name(self) -> str:
        """Get the name of the device."""
        for name, value in USB_DEVICES.items():
            if value[0] == self.vid and value[1] == self.pid:
                return name
        return "Unknown"

    @property
    def is_opened(self) -> bool:
        """Return True if device is open."""
        return self.device is not None and self._opened

    @property
    def path_str(self) -> str:
        """BLHost-friendly string representation of USB path."""
        return NXPUSBDeviceFilter.convert_usb_path(self.path)

    @property
    def path_hash(self) -> str:
        """BLHost-friendly hash of the USB path."""
        return get_hash(self.path)

    def __init__(self) -> None:
        """Initialize the USB interface object."""
        self._opened = False
        self.vid = 0
        self.pid = 0
        self.serial_number = ""
        self.vendor_name = ""
        self.product_name = ""
        self.interface_number = 0
        self.timeout = 2000
        self.path = b""
        self.device: libusbsio.LIBUSBSIO.HID_DEVICE = None

    @staticmethod
    def _encode_report(
        report_id: int, report_size: int, data: bytes, offset: int = 0
    ) -> Tuple[bytes, int]:
        """Encode the USB packet.

        :param report_id: ID of the report (see: HID_REPORT)
        :param report_size: Length of the report to send
        :param data: Data to send
        :param offset: offset within the 'data' bytes
        :return: Encoded bytes and length of the final report frame
        """
        data_len = min(len(data) - offset, report_size)
        raw_data = bytes([report_id])
        raw_data += data[offset : offset + data_len]
        raw_data += bytes([0x00] * (report_size - data_len))
        logger.debug(f"OUT[{len(raw_data)}]: {', '.join(f'{b:02X}' for b in raw_data)}")
        return raw_data, offset + data_len

    @staticmethod
    def _decode_report(raw_data: bytes) -> CmdResponse:
        """Decodes the data read on USB interface.

        :param raw_data: Data received
        :return: CmdResponse object
        """
        if not raw_data:
            raise SdpConnectionError("No data were received")
        logger.debug(f"IN [{len(raw_data)}]: {', '.join(f'{b:02X}' for b in raw_data)}")
        return CmdResponse(raw_data[0] == HID_REPORT["HAB"][0], raw_data[1:])

    def info(self) -> str:
        """Return information about the USB interface."""
        return f"{self.product_name:s} (0x{self.vid:04X}, 0x{self.pid:04X})"

    def conf(self, config: dict) -> None:
        """Set HID report data.

        :param config: parameters dictionary
        """
        if "hid_ep1" in config and "pack_size" in config:
            HID_REPORT["CMD"] = (0x01, config["pack_size"], config["hid_ep1"])
            HID_REPORT["DATA"] = (0x02, config["pack_size"], config["hid_ep1"])

    def open(self) -> None:
        """Open the interface.

        :raises SdpConnectionError: when there is no available device
        :raises SdpConnectionError: when the device can not be opened
        """
        logger.debug("Open Interface")
        try:
            if not self.device:
                raise SdpConnectionError("No device available")
            self.device.Open(self.path)
            self._opened = True
        except Exception as error:
            raise SdpConnectionError(
                f"Unable to open device '{self.path!r}' VID={self.vid} PID={self.pid} SN='{self.serial_number}'"
            ) from error

    def close(self) -> None:
        """Close the interface.

        :raises SdpConnectionError: when there is no device available
        :raises SdpConnectionError: when the device can not be closed
        """
        logger.debug("Close Interface")
        if not self.device:
            raise SdpConnectionError("No device available")
        try:
            self.device.Close()
            self._opened = False
        except OSError as error:
            raise SdpConnectionError(
                f"Unable to close device '{self.path!r}' VID={self.vid} PID={self.pid} SN='{self.serial_number}'"
            ) from error

    def write(self, packet: Union[CmdPacket, bytes]) -> None:
        """Write data on the OUT endpoint associated to the HID interfaces.

        :param packet: Data to send
        :raises SdpError: Raises an error if packet type is incorrect
        :raises SdpConnectionError: Raises an error if device is opened for writing
        :raises SdpConnectionError: Raises if device is not available
        :raises SdpConnectionError: Raises if writing to device fails
        """
        if not self.is_opened:
            raise SdpConnectionError("Device is opened for writing")

        if isinstance(packet, CmdPacket):
            report_id, report_size, hid_ep1 = HID_REPORT["CMD"]  # pylint: disable=unused-variable
            data = packet.to_bytes()
        elif isinstance(packet, (bytes, bytearray)):
            report_id, report_size, hid_ep1 = HID_REPORT["DATA"]  # pylint: disable=unused-variable
            data = packet
        else:
            raise SdpError("Packet has to be either 'CmdPacket' or 'bytes'")

        if not self.device:
            raise SdpConnectionError("No device available")
        data_index = 0
        while data_index < len(data):
            try:
                raw_data, data_index = self._encode_report(report_id, report_size, data, data_index)
                self.device.Write(raw_data, timeout_ms=self.timeout)
            except Exception as e:
                raise SdpConnectionError(str(e)) from e

    def read(self, length: Optional[int] = None) -> CmdResponse:  # pylint: disable=unused-argument
        """Read data on the IN endpoint associated to the HID interface.

        :return: Return CmdResponse object.
        :raises SdpConnectionError: Raises an error if device is opened for reading
        :raises SdpConnectionError: Raises an error if device is not available
        :raises SdpConnectionError: Raises an error if data can not be read from device
        """
        if not self.is_opened:
            raise SdpConnectionError("Device is opened for reading")

        if not self.device:
            raise SdpConnectionError("No device available")
        try:
            (raw_data, result) = self.device.Read(  # pylint: disable=unused-variable
                1024, timeout_ms=self.timeout
            )
            return self._decode_report(bytes(raw_data))
        except Exception as e:
            raise SdpConnectionError(str(e)) from e

    @staticmethod
    def enumerate(usb_device_filter: USBDeviceFilter) -> List["RawHid"]:
        """Get list of all connected devices which matches device_id.

        :param usb_device_filter: USBDeviceFilter object
        :return: List of interfaces found
        """
        devices = []
        libusbsio_logger = logging.getLogger("libusbsio")
        sio = libusbsio.usbsio(loglevel=libusbsio_logger.getEffectiveLevel())
        all_hid_devices = sio.HIDAPI_Enumerate()

        # iterate on all devices found
        for dev in all_hid_devices:
            if usb_device_filter.compare(vars(dev)) is True:
                new_device = RawHid()
                new_device.device = sio.HIDAPI_DeviceCreate()
                new_device.vid = dev["vendor_id"]
                new_device.pid = dev["product_id"]
                new_device.vendor_name = dev["manufacturer_string"]
                new_device.product_name = dev["product_string"]
                new_device.interface_number = dev["interface_number"]
                new_device.path = dev["path"]
                devices.append(new_device)

        return devices



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for NXP device description classes."""


from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

from libusbsio.libusbsio import LIBUSBSIO

from spsdk.mboot.interfaces.usb import USB_DEVICES as MB_USB_DEVICES
from spsdk.sdp.interfaces.usb import USB_DEVICES as SDP_USB_DEVICES
from spsdk.utils.misc import get_hash

# for backward-compatibility
from spsdk.utils.usbfilter import NXPUSBDeviceFilter

convert_usb_path = NXPUSBDeviceFilter.convert_usb_path


class DeviceDescription(ABC):
    """Base class for all logical devices.

    The intent is to have a generic container for providing info about devices
    of any type. Thus the class is named as 'logical', because it doesn't
    allow you to control the device in any way.

    This is just a base class and as such shouldn't be used. If you want to
    use it, create your own class inheriting from this class and redefining
    the methods listed in this class!
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({vars(self)})"

    def __str__(self) -> str:
        """Should return the string from info function."""
        return self.info()

    @abstractmethod  # pragma: no cover
    def info(self) -> str:
        """Shall return a string describing the device, e.g. Name: <name>; ID: <id>."""


class UartDeviceDescription(DeviceDescription):
    """Simple container holding information about UART device.

    This container should be used instead of any USB API related objects, as
    this container will be the same all the time compared to specific UART API
    implementations.
    """

    def __init__(self, name: Optional[str] = None, dev_type: Optional[str] = None) -> None:
        """Constructor.

        The 'dev_type' can be in general any string identifying the device type.

        :param name: COM port name
        :param dev_type: 'mboot device' or 'SDP device'
        """
        self.name = name or "Unknown port"
        self.dev_type = dev_type or "Unknown device type"

    def info(self) -> str:
        """Returns a formatted device description string.

        :return: Text information about UART device.
        """
        return f"Port: {self.name}\nType: {self.dev_type}"


class USBDeviceDescription(DeviceDescription):
    """Simple container holding information about USB device.

    This container should be used instead of any USB API related objects, as
    this container will be the same all the time compared to specific USB API
    implementations.
    """

    def __init__(
        self,
        vid: int,
        pid: int,
        path: str,
        product_string: str,
        manufacturer_string: str,
        name: str,
        serial: str,
        original_path: Optional[Union[str, bytes]] = None,
    ) -> None:
        """Constructor.

        :param vid: Vendor ID
        :param pid: Product ID
        :param product_string: Product string
        :param manufacturer_string: Manufacturer string
        :param name: Name(s) of NXP devices as defined under spsdk.mboot.interfaces.usb or spsdk.sdp.interfaces.usb
        :param serial: The serial number of device.

        See :py:func:`get_usb_device_name` function to get the name from
        VID and PID.
        See :py:func:`convert_usb_path` function to provide a proper path string.
        """
        self.vid = vid
        self.pid = pid
        self.path = path
        self.product_string = product_string
        self.manufacturer_string = manufacturer_string
        self.name = name
        self.serial = serial
        self.path_hash = get_hash(original_path) if original_path else "N/A"

    def info(self) -> str:
        """Returns a formatted device description string.

        :return: Text information of USB device.
        """
        return (
            f"{self.product_string} - {self.manufacturer_string}\n"
            f"Vendor ID: 0x{self.vid:04x}\n"
            f"Product ID: 0x{self.pid:04x}\n"
            f"Path: {self.path}\n"
            f"Path Hash: {self.path_hash}\n"
            f"Name: {self.name}\n"
            f"Serial number: {self.serial}"
        )


class SDIODeviceDescription(DeviceDescription):
    """Simple container holding information about SDIO device.

    This container should be used instead of any SDIO API related objects, as
    this container will be the same all the time compared to specific SDIO API
    implementations.
    """

    def __init__(
        self,
        vid: int,
        pid: int,
        path: str,
    ) -> None:
        """Constructor.

        :param vid: Vendor ID
        :param pid: Product ID

        See :py:func:`get_usb_device_name` function to getg the name from
        VID and PID.
        See :py:func:`convert_usb_path` function to provide a proper path string.
        """
        self.vid = vid
        self.pid = pid
        self.path = path

    def info(self) -> str:
        """Returns a formatted device description string.

        :return: Text information of SDIO device.
        """
        return (
            f"Vendor ID: 0x{self.vid:04x}\n"
            f"Product ID: 0x{self.pid:04x}\n"
            f"Path: {self.path}\n"
        )


class SIODeviceDescription(DeviceDescription):
    """Simple container holding information about LIBUSBSIO device.

    This container contains information about LIBUSBSIOdevice.
    """

    def __init__(self, info: LIBUSBSIO.HIDAPI_DEVICE_INFO_T) -> None:
        """Constructor.

        :param info: LIBUSBSIO device information class.
        """
        self._info = info
        self.pid = self._info.product_id
        self.vid = self._info.vendor_id
        self.manufacturer_string = self._info.manufacturer_string
        self.product_string = self._info.product_string
        self.path = convert_usb_path(self._info.path)
        self.serial_number = self._info.serial_number
        self.interface_number = self._info.interface_number
        self.release_number = self._info.release_number
        self.path_hash = get_hash(self._info.path)

    def info(self) -> str:
        """Returns a formatted device description string.

        :return: Text description of SIO device.
        """
        return (
            f"LIBUSBSIO - {self.manufacturer_string}, {self.product_string}\n"
            f"Vendor ID: 0x{self.vid:04x}\n"
            f"Product ID: 0x{self.pid:04x}\n"
            f"Path: {self.path}\n"
            f"Path Hash: {self.path_hash or 'N/A'}\n"
            f"Serial number: {self.serial_number}\n"
            f"Interface number: {self.interface_number}\n"
            f"Release number: {self.release_number}"
        )


def get_usb_device_name(
    vid: int, pid: int, device_names: Optional[Dict[str, Tuple[int, int]]] = None
) -> List[str]:
    """Returns 'name' device identifier based on VID/PID, from dicts.

    Searches provided dictionary for device name based on VID/PID. If the dict
    is None, the search happens on USB_DEVICES under mboot/interfaces/usb.py and
    sdphost/interfaces/usb.py

    DESIGN REMARK: this function is not part of the USBLogicalDevice, as the
    class intention is to be just a simple container. But to help the class
    to get the required inputs, this helper method has been provided.

    :param vid: Vendor ID we are interested in
    :param pid: Product ID we are interested in
    :param device_names: dict where str is device name, first int vid, second int pid

    :return: list containing device names with corresponding VID/PID
    """
    nxp_device_names = []
    if device_names is None:
        for dname, vid_pid in MB_USB_DEVICES.items():
            if vid_pid[0] == vid and vid_pid[1] == pid:
                nxp_device_names.append(dname)

        for dname, vid_pid in SDP_USB_DEVICES.items():
            if vid_pid[0] == vid and vid_pid[1] == pid:
                nxp_device_names.append(dname)
    else:
        for dname, vid_pid in device_names.items():
            if vid_pid[0] == vid and vid_pid[1] == pid:
                nxp_device_names.append(dname)

    return nxp_device_names



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Sections within SBfile."""

from struct import unpack_from
from typing import Iterator, List, Optional

from spsdk import SPSDKError
from spsdk.utils.crypto import CertBlockV2, Counter
from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.crypto.common import calc_cypher_block_count, crypto_backend

from .commands import CmdBaseClass, CmdHeader, EnumCmdTag, EnumSectionFlag, parse_command

########################################################################################################################
# Boot Image Sections
########################################################################################################################


class BootSectionV2(BaseClass):
    """Boot Section V2."""

    HMAC_SIZE = 32

    @property
    def uid(self) -> int:
        """Boot Section UID."""
        return self._header.address

    @uid.setter
    def uid(self, value: int) -> None:
        self._header.address = value

    @property
    def is_last(self) -> bool:
        """Check whether the section is the last one."""
        return self._header.flags & EnumSectionFlag.LAST_SECT != 0

    @is_last.setter
    def is_last(self, value: bool) -> None:
        assert isinstance(value, bool)
        self._header.flags = EnumSectionFlag.BOOTABLE
        if value:
            self._header.flags |= EnumSectionFlag.LAST_SECT

    @property
    def hmac_count(self) -> int:
        """Number of HMACs."""
        raw_size = 0
        hmac_count = 0
        for cmd in self._commands:
            raw_size += cmd.raw_size
        if raw_size > 0:
            block_count = (raw_size + 15) // 16
            hmac_count = self._hmac_count if block_count >= self._hmac_count else block_count
        return hmac_count

    @property
    def raw_size(self) -> int:
        """Raw size of section."""
        size = CmdHeader.SIZE + self.HMAC_SIZE
        size += self.hmac_count * self.HMAC_SIZE
        for cmd in self._commands:
            size += cmd.raw_size
        if size % 16:
            size += 16 - (size % 16)
        return size

    def __init__(self, uid: int, *commands: CmdBaseClass, hmac_count: int = 1) -> None:
        """Initialize BootSectionV2.

        :param uid: section unique identification
        :param commands: List of commands
        :param hmac_count: The number of HMAC entries
        """
        self._header = CmdHeader(EnumCmdTag.TAG, EnumSectionFlag.BOOTABLE)
        self._commands: List[CmdBaseClass] = []
        self._hmac_count = hmac_count
        for cmd in commands:
            self.append(cmd)
        # Initialize HMAC count
        if not isinstance(self._hmac_count, int) or self._hmac_count == 0:
            self._hmac_count = 1
        # section UID
        self.uid = uid

    def __str__(self) -> str:
        return f"BootSectionV2: {len(self)} commands."

    def __len__(self) -> int:
        return len(self._commands)

    def __getitem__(self, key: int) -> CmdBaseClass:
        return self._commands[key]

    def __setitem__(self, key: int, value: CmdBaseClass) -> None:
        self._commands[key] = value

    def __iter__(self) -> Iterator[CmdBaseClass]:
        return self._commands.__iter__()

    def append(self, cmd: CmdBaseClass) -> None:
        """Add command to section."""
        assert isinstance(cmd, CmdBaseClass)
        self._commands.append(cmd)

    def info(self) -> str:
        """Get object info."""
        nfo = ""
        for index, cmd in enumerate(self._commands):
            nfo += f" {index}) {cmd.info()}\n"
        return nfo

    # pylint: disable=too-many-locals
    def export(
        self,
        dek: bytes = b"",
        mac: bytes = b"",
        counter: Optional[Counter] = None,
        dbg_info: Optional[List[str]] = None,
    ) -> bytes:
        """Serialize Boot Section object.

        :param dek: The DEK value in bytes (required)
        :param mac: The MAC value in bytes (required)
        :param counter: The counter object (required)
        :param dbg_info: Optional[List[str]] optional list to export debug information about content in text format
        :return: exported bytes
        :raises SPSDKError: raised when dek, mac, counter have invalid format
        """
        cmd_dbg_info: Optional[List[str]] = None
        if dbg_info is not None:
            dbg_info.append("[bootable_section]")
            cmd_dbg_info = []
            cmd_dbg_info.append("[commands]")
        if not isinstance(dek, bytes):
            raise SPSDKError("Invalid type of dek, should be bytes")
        if not isinstance(mac, bytes):
            raise SPSDKError("Invalid type of mac, should be bytes")
        if not isinstance(counter, Counter):
            raise SPSDKError("Invalid type of counter")
        # Export commands
        commands_data = b""
        for cmd in self._commands:
            cmd_data = cmd.export()
            commands_data += cmd_data
            if cmd_dbg_info is not None:
                cmd_dbg_info.append(f"[command:{type(cmd).__name__}]")
                cmd_dbg_info.append(cmd_data.hex())
        if len(commands_data) % 16:
            commands_data += b"\x00" * (16 - (len(commands_data) % 16))
        # Encrypt header
        self._header.data = self.hmac_count
        self._header.count = len(commands_data) // 16
        encrypted_header = crypto_backend().aes_ctr_encrypt(
            dek, self._header.export(), counter.value
        )
        hmac_data = crypto_backend().hmac(mac, encrypted_header)
        counter.increment(1 + (self.hmac_count + 1) * 2)
        if dbg_info:
            dbg_info.append("[header]")
            dbg_info.append(self._header.export().hex())
            dbg_info.append("[encrypted_header]")
            dbg_info.append(encrypted_header.hex())
            dbg_info.append("[hmac]")
            dbg_info.append(hmac_data.hex())
            if cmd_dbg_info:
                dbg_info.extend(cmd_dbg_info)
        # Encrypt commands
        encrypted_commands = b""
        for index in range(0, len(commands_data), 16):
            encrypted_block = crypto_backend().aes_ctr_encrypt(
                dek, commands_data[index : index + 16], counter.value
            )
            encrypted_commands += encrypted_block
            counter.increment()
        # Calculate HMAC of commands
        index = 0
        hmac_count = self._header.data
        block_size = (self._header.count // hmac_count) * 16
        while hmac_count > 0:
            enc_block = (
                encrypted_commands[index:]
                if hmac_count == 1
                else encrypted_commands[index : index + block_size]
            )
            hmac_data += crypto_backend().hmac(mac, enc_block)
            hmac_count -= 1
            index += len(enc_block)
        return encrypted_header + hmac_data + encrypted_commands

    # pylint: disable=too-many-locals
    @classmethod
    def parse(
        cls,
        data: bytes,
        offset: int = 0,
        plain_sect: bool = False,
        dek: bytes = b"",
        mac: bytes = b"",
        counter: Optional[Counter] = None,
    ) -> "BootSectionV2":
        """Parse Boot Section from bytes.

        :param data: Raw data of parsed image
        :param offset: The offset of input data
        :param plain_sect: If the sections are not encrypted; It is used for debugging only, not supported by ROM code
        :param dek: The DEK value in bytes (required)
        :param mac: The MAC value in bytes (required)
        :param counter: The counter object (required)
        :return: exported bytes
        :raises SPSDKError: raised when dek, mac, counter have invalid format
        """
        if not isinstance(dek, bytes):
            raise SPSDKError("Invalid type of dek, should be bytes")
        if not isinstance(mac, bytes):
            raise SPSDKError("Invalid type of mac, should be bytes")
        if not isinstance(counter, Counter):
            raise SPSDKError("Invalid type of counter")
        # Get Header specific data
        header_encrypted = data[offset : offset + CmdHeader.SIZE]
        header_hmac_data = data[offset + CmdHeader.SIZE : offset + CmdHeader.SIZE + cls.HMAC_SIZE]
        offset += CmdHeader.SIZE + cls.HMAC_SIZE
        # Check header HMAC
        if header_hmac_data != crypto_backend().hmac(mac, header_encrypted):
            raise SPSDKError("Invalid header HMAC")
        # Decrypt header
        header_decrypted = crypto_backend().aes_ctr_decrypt(dek, header_encrypted, counter.value)
        counter.increment()
        # Parse header
        header = CmdHeader.parse(header_decrypted)
        counter.increment((header.data + 1) * 2)
        # Get HMAC data
        hmac_data = data[offset : offset + (cls.HMAC_SIZE * header.data)]
        offset += cls.HMAC_SIZE * header.data
        encrypted_commands = data[offset : offset + (header.count * 16)]
        # Check HMAC
        hmac_index = 0
        hmac_count = header.data
        block_size = (header.count // hmac_count) * 16
        section_size = header.count * 16
        while hmac_count > 0:
            if hmac_count == 1:
                block_size = section_size
            hmac_block = crypto_backend().hmac(mac, data[offset : offset + block_size])
            if hmac_block != hmac_data[hmac_index : hmac_index + cls.HMAC_SIZE]:
                raise SPSDKError("HMAC failed")
            hmac_count -= 1
            hmac_index += cls.HMAC_SIZE
            section_size -= block_size
            offset += block_size
        # Decrypt commands
        decrypted_commands = b""
        for hmac_index in range(0, len(encrypted_commands), 16):
            encr_block = encrypted_commands[hmac_index : hmac_index + 16]
            decrypted_block = (
                encr_block
                if plain_sect
                else crypto_backend().aes_ctr_decrypt(dek, encr_block, counter.value)
            )
            decrypted_commands += decrypted_block
            counter.increment()
        # ...
        cmd_offset = 0
        obj = cls(header.address, hmac_count=header.data)
        while cmd_offset < len(decrypted_commands):
            cmd_obj = parse_command(decrypted_commands, cmd_offset)
            cmd_offset += cmd_obj.raw_size
            obj.append(cmd_obj)
        return obj


class CertSectionV2(BaseClass):
    """Certificate Section V2 class."""

    HMAC_SIZE = 32
    SECT_MARK = unpack_from("<L", b"sign")[0]

    @property
    def cert_block(self) -> CertBlockV2:
        """Return certification block."""
        return self._cert_block

    @property
    def raw_size(self) -> int:
        """Calculate raw size of section."""
        # Section header size
        size = CmdHeader.SIZE
        # Header HMAC 32 bytes + Certificate block HMAC 32 bytes
        size += self.HMAC_SIZE * 2
        # Certificate block size in bytes
        size += self.cert_block.raw_size
        return size

    def __init__(self, cert_block: CertBlockV2):
        """Initialize CertBlockV2."""
        assert isinstance(cert_block, CertBlockV2)
        self._header = CmdHeader(
            EnumCmdTag.TAG, EnumSectionFlag.CLEARTEXT | EnumSectionFlag.LAST_SECT
        )
        self._header.address = self.SECT_MARK
        self._header.count = cert_block.raw_size // 16
        self._header.data = 1
        self._cert_block = cert_block

    def __str__(self) -> str:
        return f"CertSectionV2: Length={self._header.count * 16}"

    def info(self) -> str:
        """Get object info."""
        return self.cert_block.info()

    def export(
        self, dek: bytes = b"", mac: bytes = b"", counter: Optional[Counter] = None
    ) -> bytes:
        """Serialize Certificate Section object.

        :param dek: The DEK value in bytes (required)
        :param mac: The MAC value in bytes (required)
        :param counter: The counter object (required)
        :return: exported bytes
        :raises SPSDKError: raised when dek, mac, counter have invalid format
        :raises SPSDKError: Raised size of exported bytes is invalid
        """
        if not isinstance(dek, bytes):
            raise SPSDKError("DEK value is not in bytes")
        if not isinstance(mac, bytes):
            raise SPSDKError("MAC value is not in bytes")
        if not isinstance(counter, Counter):
            raise SPSDKError("Counter value is not incorrect")
        # Prepare Header data
        header_data = self._header.export()
        header_encrypted = crypto_backend().aes_ctr_encrypt(dek, header_data, counter.value)
        # counter.increment()
        # Prepare Certificate Block data
        body_data = self.cert_block.export()
        # Prepare HMAC data
        hmac_data = crypto_backend().hmac(mac, header_encrypted)
        hmac_data += crypto_backend().hmac(mac, body_data)
        result = header_encrypted + hmac_data + body_data
        if len(result) != self.raw_size:
            raise SPSDKError("Invalid size")
        return result

    @classmethod
    def parse(
        cls,
        data: bytes,
        offset: int = 0,
        dek: bytes = b"",
        mac: bytes = b"",
        counter: Optional[Counter] = None,
    ) -> "CertSectionV2":
        """Parse Certificate Section from bytes array.

        :param data: Raw data of parsed image
        :param offset: The offset of input data
        :param dek: The DEK value in bytes (required)
        :param mac: The MAC value in bytes (required)
        :param counter: The counter object (required)
        :return: parsed cert section v2 object
        :raises SPSDKError: Raised when dek, mac, counter are not valid
        :raises SPSDKError: Raised when there is invalid header HMAC, TAG, FLAGS, Mark
        :raises SPSDKError: Raised when there is invalid certificate block HMAC
        """
        if not isinstance(dek, bytes):
            raise SPSDKError("DEK value has invalid format")
        if not isinstance(mac, bytes):
            raise SPSDKError("MAC value has invalid format")
        if not isinstance(counter, Counter):
            raise SPSDKError("Counter value has invalid format")
        index = offset
        header_encrypted = data[index : index + CmdHeader.SIZE]
        index += CmdHeader.SIZE
        header_hmac = data[index : index + cls.HMAC_SIZE]
        index += cls.HMAC_SIZE
        cert_block_hmac = data[index : index + cls.HMAC_SIZE]
        index += cls.HMAC_SIZE
        if header_hmac != crypto_backend().hmac(mac, header_encrypted):
            raise SPSDKError("Invalid Header HMAC")
        header_encrypted = crypto_backend().aes_ctr_decrypt(dek, header_encrypted, counter.value)
        header = CmdHeader.parse(header_encrypted)
        if header.tag != EnumCmdTag.TAG:
            raise SPSDKError(f"Invalid Header TAG: 0x{header.tag:02X}")
        if header.flags != (EnumSectionFlag.CLEARTEXT | EnumSectionFlag.LAST_SECT):
            raise SPSDKError(f"Invalid Header FLAGS: 0x{header.flags:02X}")
        if header.address != cls.SECT_MARK:
            raise SPSDKError(f"Invalid Section Mark: 0x{header.address:08X}")
        # Parse Certificate Block
        cert_block = CertBlockV2.parse(data, index)
        if cert_block_hmac != crypto_backend().hmac(mac, data[index : index + cert_block.raw_size]):
            raise SPSDKError("Invalid Certificate Block HMAC")
        index += cert_block.raw_size
        cert_section_obj = cls(cert_block)
        counter.increment(calc_cypher_block_count(index - offset))
        return cert_section_obj



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Lexer for command (BD) files used by parser."""

from typing import List, Union

from sly import Lexer
from sly.lex import Token


# pylint: disable=undefined-variable,invalid-name,no-self-use
# undefined-variable : the lexer uses '_' as a decorator, which throws undefined
#   variable error. We can't do much with it.
# invalid-name : tokens are defined as upper case. However this violates the
#   snake cae naming style. We can't do much, as this is required by the lexer.
# no-self-use : the public methods must be defined as class methods although
#   the self is not used at all.
class Variable:
    """Class representing a variable in command file."""

    def __init__(self, name: str, token: str, value: Union[str, int, float]) -> None:
        """Initializer.

        :param name: name of identifier (variable)
        :param token: type of variable (option, constant etc.)
        :param value: the content of the variable
        """
        self.name = name
        self.t = token
        self.value = value

    def __str__(self) -> str:
        """Returns a string with variable info.

        i.e.
        "<var_name>, <var_type>, <var_value>"

        :return: string
        """
        return f"{self.name}, {self.t}, {self.value}"


class BDLexer(Lexer):
    """Lexer for bd files."""

    def __init__(self) -> None:
        """Initializer."""
        self._sources: List[Variable] = []

    def cleanup(self) -> None:
        """Resets the lexers internals into initial state."""
        self._sources.clear()

    def add_source(self, source: Variable) -> None:
        """Append an identifier of source type into list.

        :param source: identifier defined under sources block in BD file
        """
        self._sources.append(source)

    # List of reserved keywords
    reserved = {
        "call": "CALL",
        "constants": "CONSTANTS",
        "extern": "EXTERN",
        "erase": "ERASE",
        "false": "FALSE",
        "filters": "FILTERS",
        "from": "FROM",
        "jump": "JUMP",
        "load": "LOAD",
        "mode": "MODE",
        "else": "ELSE",
        "info": "INFO",
        "error": "ERROR",
        "qspi": "QSPI",
        "enable": "ENABLE",
        "keywrap": "KEYWRAP",
        "keystore_to_nv": "KEYSTORE_TO_NV",
        "keystore_from_nv": "KEYSTORE_FROM_NV",
        "all": "ALL",
        "no": "NO",
        "options": "OPTIONS",
        "raw": "RAW",
        "section": "SECTION",
        "sources": "SOURCES",
        "switch": "SWITCH",
        "true": "TRUE",
        "yes": "YES",
        "if": "IF",
        "defined": "DEFINED",
        "warning": "WARNING",
        "sizeof": "SIZEOF",
        "unsecure": "UNSECURE",
        "jump_sp": "JUMP_SP",
        "keyblob": "KEYBLOB",
        "reset": "RESET",
        "encrypt": "ENCRYPT",
        "version_check": "VERSION_CHECK",
        "sec": "SEC",
        "nsec": "NSEC",
    }

    # List of token names. This is always required
    tokens = [
        "COMMENT",
        "IDENT",
        "SOURCE_NAME",
        "BINARY_BLOB",
        "INT_LITERAL",
        "STRING_LITERAL",
        "RANGE",
        "ASSIGN",
        "INT_SIZE",
        "SECTION_NAME",
        #'SYMBOL_REF', replaced with a non-terminal symbol_ref
        # Operators (+,-,*,/,%,|,&,~,^,<<,>>, ||, &&, !, <, <=, >, >=, ==, !=)
        "PLUS",
        "MINUS",
        "TIMES",
        "DIVIDE",
        "MOD",
        "OR",
        "AND",
        "NOT",
        "XOR",
        "LSHIFT",
        "RSHIFT",
        "LOR",
        "LAND",
        "LNOT",
        "LT",
        "LE",
        "GT",
        "GE",
        "EQ",
        "NE",
        # Delimiters ( ) { } , . ; :
        "LPAREN",
        "RPAREN",
        "LBRACE",
        "RBRACE",
        "COMMA",
        "PERIOD",
        "SEMI",
        "COLON",
        # Special characters
        "QUESTIONMARK",
        "DOLLAR",
    ] + list(reserved.values())

    literals = {"@"}

    # A regular expression rules with some action code
    # The order of these functions MATTER!!! Make sure you know what you are
    # doing, when changing the order of function declarations!!!
    @_(r"(//.*)|(/\*(.|\s)*?\*/)|(\#.*)")  # type: ignore
    def COMMENT(self, token: Token) -> None:
        """Token rule to detect comments (including multiline).

        Allowed comments are C/C++ like comments '/* */', '//' and bash-like
        comments starting with '#'.

        :param token: token matching a comment
        """
        # Multiline comments are counted as a single line. This causes us troubles
        # in t_newline(), which treats the multiline comment as a single line causing
        # a mismatch in the final line position.
        # From this perspective we increment the linenumber here by the total
        # number of lines - 1 (the subtracted 1 gets counted byt t_newline)
        self.lineno += len(token.value.split("\n")) - 1

    # It's not possible to detect INT_SIZE token while whitespaces are present between period and
    # letter in real use case, because of regex engine limitation in positive lookbehind.
    @_(r"(?<=(\d|[0-9a-fA-F])\.)[ \t]*[whb]")  # type: ignore
    def INT_SIZE(self, token: Token) -> Token:
        """Token rule to detect numbers appended with w/h/b.

        Example:
        my_number = 4.b
        my_number = 1.h
        my_number = 3.w

        The w/h/b defines size (Byte, Halfword, Word). This should be taken into
        account during number computation.

        :param token: token matching int size

        :return: Token representing the size of int literal
        """
        return token

    @_(r"[_a-zA-Z][_a-zA-Z0-9]*")  # type: ignore
    def IDENT(self, token: Token) -> Token:
        """Token rule to detect identifiers.

        A valid identifier can start either with underscore or a letter followed
        by any numbers of underscores, letters and numbers.

        If the name of an identifier is from the set of reserved keywords, the
        token type is replaced with the keyword name, otherwise the token is
        of type 'IDENT'.
        Values of type TRUE/YES, FALSE/NO are replaces by 1 or 0 respectively.

        :param token: token matching an identifier pattern
        :return: Token representing identifier
        """
        # it may happen that we find an identifier, which is a keyword, in such
        # a case remap the type from IDENT to reserved word (i.e. keyword)
        token_type = self.reserved.get(token.value, "IDENT")
        if token_type in ["TRUE", "YES"]:
            token.type = "INT_LITERAL"
            token.value = 1
        elif token_type in ["FALSE", "NO"]:
            token.type = "INT_LITERAL"
            token.value = 0
        else:
            token.type = token_type
            # check, whether the identifier is under sources, in such case
            # change the type to SOURCE_NAME
            for source in self._sources:
                if source.name == token.value:
                    token.type = "SOURCE_NAME"
                    break
        return token

    @_(r"\b([0-9]+[K]?|0[xX][0-9a-fA-F]+)\b|'.*'")  # type: ignore
    def INT_LITERAL(self, token: Token) -> Token:
        """Token rule to detect integer literals.

        An int literal may be represented as a number in decimal form appended
        with a 'K' or number in hexadecimal form.

        Example:
        1024
        1K # same as above
        -256
        0x25

        Lexer converts the detected string into a number. String literals
        appended with 'K' are multiplied by 1024.

        :param token: token matching integer literal pattern
        :return: Token representing integer literal
        """
        number = token.value
        if number[0] == "'" and number[-1] == "'":
            # transform 'dude' into '0x64756465'
            number = "0x" + bytearray(number[1:-1], "utf-8").hex()
            number = int(number, 0)
        elif number[-1] == "K":
            number = int(number[:-1], 0) * 1024
        else:
            number = int(number, 0)

        token.value = number
        return token

    @_(r"\$[\w\.\*\?\-\^\[\]]+")  # type: ignore
    def SECTION_NAME(self, token: Token) -> Token:
        """Token rule to detect section names.

        Section names start with a dollar sign ($) glob-type expression that
        can match any number of ELF sections.

        Example:
        $section_[ab]
        $math*

        :param token: token matching section name pattern
        :return: Token representing section name
        """
        return token

    @_(r"\{\{([0-9a-fA-F]{2}| )+\}\}")  # type: ignore
    def BINARY_BLOB(self, token: Token) -> Token:
        """Token rule to detect binary blob.

        A binary blob is a sequence of hexadecimal bytes in double curly braces.

        Example:
        {{aa bb cc 1F 3C}}

        :param token: token matching binary blob pattern
        :return: Token representing binary blob
        """
        # return just the content between braces
        value = token.value[2:-2]

        token.value = "".join(value.split())
        return token

    # A string containing ignored characters (spaces and tabs)
    ignore = " \t"

    @_(r"\n")  # type: ignore
    def newline(self, token: Token) -> None:
        """Token rule to detect new lines.

        On new line character the line number count is incremented.

        :param token: token matching new line character
        """
        self.lineno += len(token.value)

    # Operators regular expressions
    PLUS = r"\+"
    MINUS = r"-"
    TIMES = r"\*"
    DIVIDE = r"/"
    MOD = r"%"
    NOT = r"~"
    XOR = r"\^"
    LSHIFT = r"<<"
    RSHIFT = r">>"
    LOR = r"\|\|"
    OR = r"\|"
    LAND = r"&&"
    AND = r"&"
    LE = r"<="
    LT = r"<"
    GE = r">="
    GT = r">"
    EQ = r"=="
    NE = r"!="
    LNOT = r"!"

    # Tokens regular expressions
    STRING_LITERAL = r"\".*\""
    RANGE = r"\.\."

    # Assignment operator regular expressions
    ASSIGN = r"="

    # Delimiters regular expressions
    LPAREN = r"\("
    RPAREN = r"\)"
    LBRACE = r"\{"
    RBRACE = r"\}"
    COMMA = r","
    PERIOD = r"\."
    SEMI = r";"
    COLON = r":"

    # Special characters
    QUESTIONMARK = r"\?"
    DOLLAR = r"\$"

    # Error handling rule
    def error(self, t: Token) -> Token:
        """Token error handler.

        The lexing index is incremented so lexing can continue, however, an
        error token is returned. The token contains the whole text starting
        with the detected error.

        :param t: invalid token.
        :return: the invalid token.
        """
        self.index += 1
        t.value = t.value[0]
        return t



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""HAB Binary Image extension."""

from enum import Enum
from typing import Optional

from spsdk.exceptions import SPSDKKeyError
from spsdk.image import segments
from spsdk.image.images import BootImgRT
from spsdk.utils.images import BinaryImage
from spsdk.utils.misc import BinaryPattern, align_block, find_first


class HabSegment(str, Enum):
    """Enum definition for HAB segments."""

    IVT = "ivt"
    BDT = "bdt"
    DCD = "dcd"
    XMCD = "xmcd"
    CSF = "csf"
    APP = "app"


class HabBinaryImage(BinaryImage):
    """HAB binary image."""

    IVT_OFFSET = 0
    BDT_OFFSET = IVT_OFFSET + segments.SegIVT2.SIZE
    DCD_OFFSET = BDT_OFFSET + BootImgRT.BDT_SIZE
    XMCD_OFFSET = IVT_OFFSET + BootImgRT.XMCD_IVT_OFFSET

    CSF_SIZE = 0x2000
    KEYBLOB_SIZE = 0x200

    def __init__(self) -> None:
        """HAB Binary Image Constructor."""
        super().__init__(name="HAB", size=0, pattern=BinaryPattern("zeros"))

    def add_hab_segment(
        self,
        segment_name: HabSegment,
        binary: bytes,
        offset_override: Optional[int] = None,
    ) -> BinaryImage:
        """Create binary image and add it into parent image.

        :param segment_name: Segment to be added
        :param binary: Segment binary
        :param offset_override: Segment offset in the parent image
        :return: Created bootable image
        """
        segment_defintions = {
            HabSegment.IVT: (segments.SegIVT2.SIZE, self.IVT_OFFSET),
            HabSegment.BDT: (segments.SegBDT.SIZE, self.BDT_OFFSET),
            HabSegment.DCD: (None, self.DCD_OFFSET),
            HabSegment.XMCD: (None, self.XMCD_OFFSET),
            HabSegment.CSF: (BootImgRT.CSF_SIZE, None),
        }
        size, offset = segment_defintions.get(segment_name, (None, None))
        if size is None:
            size = len(binary)
        if offset_override is not None:
            offset = offset_override
        assert offset is not None
        image = BinaryImage(
            name=segment_name.value,
            size=size,
            offset=offset,
            binary=binary,
            parent=self,
        )
        self.add_image(image)
        return image

    def get_hab_segment(self, segment_name: HabSegment) -> BinaryImage:
        """Get HAB segment.

        :param segment_name: Segment to be added
        :raises SPSDKKeyError: If HAB segment not found.
        :return: Segment as binary image
        """
        seg = find_first(self.sub_images, lambda x: x.name == segment_name.value)
        if seg is None:
            raise SPSDKKeyError(f"Segment with name {segment_name} does not exist.")
        return seg

    def align_segment(self, segment_name: HabSegment, alignment: int = 16) -> None:
        """Align HAB segment.

        :param segment_name: Segment to be aligned
        :param alignment: Alignement length
        """
        seg = self.get_hab_segment(segment_name)
        assert seg.binary
        aligned_seg = align_block(seg.binary, alignment)
        seg.binary = aligned_seg
        seg.size = len(aligned_seg)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module to manage used databases in SPSDK."""


import dataclasses
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from spsdk.exceptions import SPSDKError, SPSDKTypeError, SPSDKValueError
from spsdk.utils.misc import find_first, load_configuration


@dataclass
class Revision:
    """Revision dataclass represents a single device revision."""

    name: str
    is_latest: bool
    data_file: Optional[str] = None
    attributes: dict = dataclasses.field(default_factory=dict)

    @staticmethod
    def load(name: str, revision: dict, is_latest: bool = False) -> "Revision":
        """Loads the revision from dictionary.

        :param name: The revision name.
        :param revision: Revision data.
        :param is_latest: Is latest device revision.
        :raises SPSDKTypeError: In case the revision is not a dictionary type
        :return: The Revision object.
        """
        if not isinstance(revision, dict):
            raise SPSDKTypeError(f"Revision {name} must be a dict type.")
        rev = Revision(name=name, is_latest=is_latest)
        rev.data_file = revision.get("data_file")
        rev.attributes = revision.get("attributes", {})
        return rev


class Revisions(List[Revision]):
    """List of device revisions."""

    @property
    def revision_names(self) -> List[str]:
        """Get list of revisions.

        :return: List of all supported device version.
        """
        return [rev.name for rev in self]

    def get(self, name: Optional[str] = None) -> Revision:
        """Get the revision by its name.

        If name is not specified, or equal to 'latest', then the latest revision is returned.

        :param name: The revision name.
        :return: The Revision object.
        """
        if name is None or name == "latest":
            return self.get_latest()
        return self.get_by_name(name)

    def get_by_name(self, name: str) -> Revision:
        """Get the required revision.

        :param name: Required revision name
        :raises SPSDKValueError: Incase of invalid device or revision value.
        :return: The Revision object.
        """
        revision = find_first(self, lambda rev: rev.name == name)
        if not revision:
            raise SPSDKValueError(f"Requested revision {revision} is not supported.")
        return revision

    def get_latest(self) -> Revision:
        """Get latest revision for device.

        :raises SPSDKValueError: Incase of there is no latest revision defined.
        :return: The Revision object.
        """
        revision = find_first(self, lambda rev: rev.is_latest)
        if not revision:
            raise SPSDKValueError("No latest revision has been defined.")
        return revision

    @staticmethod
    def load(revisions: dict, latest: Optional[str] = None) -> "Revisions":
        """Loads the revisions list from dictionary.

        :param revisions: Revisions data.
        :param latest: Name of latest revision.
        :raises SPSDKTypeError: In case the revisions parameter is not a dictionary type
        :return: The Revisions object.
        """
        if not isinstance(revisions, dict):
            raise SPSDKTypeError(f"Revisions must be a dictionary type, not {type(revisions)}.")
        return Revisions(
            Revision.load(name, val, is_latest=name == latest) for name, val in revisions.items()
        )


@dataclass
class Device:
    """Device dataclass represents a single device."""

    name: str
    device_alias: Optional[str] = None
    revisions: Revisions = dataclasses.field(default_factory=Revisions)
    attributes: dict = dataclasses.field(default_factory=dict)

    @staticmethod
    def load(name: str, device: dict) -> "Device":
        """Loads the device from dictionary.

        :param name: The name of device.
        :param device: Device data.
        :return: The Device object.
        """
        dev = Device(name=name)
        if "device_alias" in device:
            dev.device_alias = device["device_alias"]
        dev.revisions = Revisions.load(
            device.get("revisions", {}), latest=device.get("latest", None)
        )
        dev.attributes = device.get("attributes", {})
        return dev


class Devices(List[Device]):
    """List of devices."""

    @property
    def device_names(self) -> List[str]:
        """Get the list of all device names."""
        devices = [dev.name for dev in self]
        devices.sort()
        return devices

    def get_by_name(self, name: str) -> Device:
        """Return database device structure.

        :param name: String Key with device name.
        :raises SPSDKValueError: In case the device with given name does not exist
        :return: Dictionary device configuration structure or None:
        """
        dev = find_first(self, lambda dev: dev.name == name)
        if not dev:
            raise SPSDKValueError(f"The device with name {name} is not in the database.")
        if dev.device_alias:
            return self.get_by_name(dev.device_alias)
        return dev

    @staticmethod
    def load(devices: dict) -> "Devices":
        """Loads the device from dictionary.

        :param devices: Devices data.
        :return: The Devices object.
        """
        return Devices(Device.load(name, val) for name, val in devices.items())

    @staticmethod
    def load_from_file(path: str) -> "Devices":
        """Loads the device from database file.

        :param path: path to database file.
        :return: The Devices object.
        """
        config: Dict[str, Any] = load_configuration(path)
        return Devices.load(config["devices"])


class Database:
    """Class that helps manage used databases in SPSDK."""

    def __init__(self, path: str, index: Optional[int] = None) -> None:
        """Register Configuration class constructor.

        :param path: The path to configuration JSON file.
        :param index: Values with {index} will be replaced with index value
        """
        self.path = path
        self.index = index
        config: Dict[str, Any] = load_configuration(path)
        try:
            self._devices = Devices.load(config["devices"])
        except (SPSDKValueError, SPSDKTypeError) as exc:
            if exc.description:
                exc.description += f"File path: {self.path}"
            raise SPSDKError("Database can not be created") from exc
        self.attributes: dict = config.get("attributes", {})

    @classmethod
    def get_devices(cls, path: str) -> Devices:
        """Classmethod to get list of supported devices.

        :param path: Path to database file.
        :return: List of all supported devices.
        """
        return Devices.load_from_file(path)

    @property
    def devices(self) -> Devices:
        """Get the list of devices stored in the database."""
        return self._devices

    def replace_idx_value(self, value: str) -> str:
        """Replace index value if provided in the database.

        :param value: value to be replaced f-string containing index
        :return: value with replaced index
        """
        if self.index and isinstance(value, str):
            value = value.replace("{index}", str(self.index))
        if self.index and isinstance(value, list):
            for reg in value:
                reg["name"] = reg["name"].replace("{index}", str(self.index))
        return value

    def get_device_value(
        self,
        key: str,
        device: Optional[str] = None,
        revision: str = "latest",
        default: Optional[Any] = None,
    ) -> Any:
        """Return any parameter by key and replace the index if provided in DB.

        :param key: The Key of the parameter to be returned.
        :param device: The device name.
        :param revision: The revision of the silicon.
        :param default: The default Value in case that is not specified in config file.
        :return: The Value of parameter by handled Key.
        """
        if device and device in self.devices.device_names:
            dev = self.devices.get_by_name(device)
            try:
                rev = dev.revisions.get(revision)
            except SPSDKValueError:
                rev = None
            if rev and key in rev.attributes:
                return self.replace_idx_value(rev.attributes[key])
            if key in dev.attributes:
                return self.replace_idx_value(dev.attributes[key])

        return self.replace_idx_value(self.attributes.get(key, default))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains support for Shadow Registers Tool."""
from .shadowregs import ShadowRegisters, enable_debug



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause


"""Parser of BD configuration."""

from collections import UserDict
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from spsdk.exceptions import SPSDKKeyError


@dataclass
class ImageConfig:
    """Represent parsed image configuration including options and sections."""

    elf_file: str
    options: "ConfigOptions"
    sections: List["SectionConfig"]

    @staticmethod
    def parse(data: Dict[str, Any]) -> "ImageConfig":
        """Parse config from dictionary.

        :param data: Configuration data to be parsed.
        """
        options = ConfigOptions.parse(data["options"])
        sections: List = []
        for section in data["sections"]:
            sections.append(SectionConfig.parse(section))
        return ImageConfig(elf_file=data["sources"]["elfFile"], options=options, sections=sections)

    def get_section(self, section_index: int) -> Optional["SectionConfig"]:
        """Get config section by section id.

        :param section_index: Section with index to be retrieved
        """
        for section in self.sections:
            if section.index == section_index:
                return section
        return None


@dataclass
class ConfigOptions:
    """Dataclass holding configuration options."""

    flags: int
    start_address: int
    ivt_offset: int
    initial_load_size: int
    entrypoint_address: int
    signature_timestamp: Optional[datetime] = None
    dcd_file_path: Optional[str] = None
    xmcd_file_path: Optional[str] = None

    _FIELD_MAPPING = {
        "flags": "flags",
        "startaddress": "start_address",
        "ivtoffset": "ivt_offset",
        "initialloadsize": "initial_load_size",
        "entrypointaddress": "entrypoint_address",
        "signaturetimestamp": "signature_timestamp",
        "dcdfilepath": "dcd_file_path",
        "xmcdfilepath": "xmcd_file_path",
    }

    @staticmethod
    def parse(options: Dict[str, Any]) -> "ConfigOptions":
        """Parse config options from dictionary.

        :param options: Options to be parsed
        :raise SPSDKKeyError: If unexpected key is present
        """
        params = {}
        for opt in options:
            if opt.lower() not in ConfigOptions._FIELD_MAPPING:
                raise SPSDKKeyError(f"Unexpected option field {opt}")
            value = options[opt]
            if opt.lower() == "signaturetimestamp":
                value = datetime.strptime(value, "%d/%m/%Y %H:%M:%S").replace(tzinfo=timezone.utc)
            params[ConfigOptions._FIELD_MAPPING[opt.lower()]] = value
        return ConfigOptions(**params)


class CaseInsensitiveDict(UserDict):
    """Case insensitive dictionary."""

    def __getitem__(self, key: str) -> Any:
        for item in self:
            if isinstance(item, str) and item.lower() == key.lower():
                return self.data[item]
        raise SPSDKKeyError(f"The key {key} was not found.")


@dataclass
class SectionConfig:
    """Dataclass holding single section data."""

    index: int
    options: CaseInsensitiveDict

    @staticmethod
    def parse(section: Dict[str, Any]) -> "SectionConfig":
        """Parse config section from dictionary.

        :param section: Section to be parsed
        """
        index = int(section["section_id"])
        options = CaseInsensitiveDict((key, d[key]) for d in section["options"] for key in d)
        return SectionConfig(index=index, options=options)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for general utilities used by TP applications."""
import itertools
import os
from typing import Any, Callable, Dict, List, Optional, Union

import click
import colorama
import prettytable
from typing_extensions import Literal

from spsdk import SPSDKError
from spsdk.apps.utils import FC
from spsdk.tp import TP_DATA_FOLDER, TP_SCH_FILE, SPSDKTpError
from spsdk.tp.tp_intf import TpDevInterface
from spsdk.tp.tphost import TrustProvisioningHost
from spsdk.tp.utils import (
    TpIntfDescription,
    get_tp_device_class,
    get_tp_device_types,
    get_tp_target_class,
    get_tp_target_types,
    scan_tp_devices,
    scan_tp_targets,
    single_tp_device_adapter,
    single_tp_target_adapter,
)
from spsdk.utils.misc import find_file, load_binary, load_configuration
from spsdk.utils.schema_validator import ValidationSchemas, check_config


class TPBaseConfig:
    """Base class for TP app configs."""

    SCHEMA_FILE_DIR = TP_DATA_FOLDER
    SCHEMA_MEMBERS: List[str] = []

    def __init__(self, config_data: dict, config_dir: Optional[str] = None) -> None:
        """Initialize the basic configuration.

        :param config_data: Initial configuration data
        :param config_dir: Path to configuration file
        """
        self.config_data = config_data
        self.config_dir = config_dir

    def _validate(self) -> None:
        """Validate configuration data using appropriate validation schema."""
        schema_cfg = ValidationSchemas.get_schema_file(TP_SCH_FILE)

        # Get this app type scheme pieces
        sch_list = [schema_cfg[x] for x in self.SCHEMA_MEMBERS]
        # First check common settings
        check_config(
            config=self.config_data,
            schemas=sch_list,
            search_paths=[self.config_dir] if self.config_dir else None,
        )

        # Add device schemas
        if "device" in self.SCHEMA_MEMBERS:
            sch_list.extend(
                get_tp_device_class(self.config_data["tp_device"]).get_validation_schemas()
            )
        # Optionally add also target schemas
        if "target" in self.SCHEMA_MEMBERS:
            sch_list.extend(
                get_tp_target_class(self.config_data["tp_target"]).get_validation_schemas()
            )

        check_config(
            config=self.config_data,
            schemas=sch_list,
            search_paths=[self.config_dir] if self.config_dir else None,
        )

    @staticmethod
    def _get_config_data(config_file_path: Optional[str] = None) -> Dict[str, Any]:
        """Setup initial configuration data."""
        if config_file_path:
            return load_configuration(config_file_path)
        return {"timeout": 60}

    @property
    def tp_device(self) -> str:
        """Trust Provisioning device."""
        return self.config_data["tp_device"]

    @property
    def tp_device_parameter(self) -> Dict[str, Any]:
        """Trust Provisioning device parameters."""
        return self.config_data["tp_device_parameter"]

    @property
    def timeout(self) -> int:
        """Timeout."""
        return self.config_data.get("timeout", 60)

    @property
    def family(self) -> str:
        """Target chip family."""
        return self.config_data["family"]


class TPHostConfig(TPBaseConfig):
    """Configuration class for TPHost app."""

    SCHEMA_MEMBERS = [
        "family",
        "provisioning_firmware",
        "oem_firmware",
        "tp_timeout",
        "audit_log",
        "device",
        "target",
    ]

    def __init__(
        self,
        tp_device: Optional[str] = None,
        tp_device_parameter: Optional[List[str]] = None,
        tp_target: Optional[str] = None,
        tp_target_parameter: Optional[List[str]] = None,
        family: Optional[str] = None,
        firmware: Optional[str] = None,
        prov_firmware: Optional[str] = None,
        audit_log: Optional[str] = None,
        audit_log_key: Optional[str] = None,
        timeout: Optional[int] = None,
        config: Optional[str] = None,
    ) -> None:
        """Initialize the TPHost configuration."""
        config_data = self._get_config_data(config)
        config_dir = os.path.dirname(config) if config else None

        super().__init__(config_data=config_data, config_dir=config_dir)

        if timeout is not None:
            self.config_data["timeout"] = timeout
        if tp_device:
            self.config_data["tp_device"] = tp_device
        if tp_target:
            self.config_data["tp_target"] = tp_target
        if family:
            self.config_data["family"] = family
        if firmware:
            self.config_data["firmware"] = firmware
        if prov_firmware:
            self.config_data["prov_firmware"] = prov_firmware
        if audit_log:
            self.config_data["audit_log"] = audit_log
        if audit_log_key:
            self.config_data["audit_log_key"] = audit_log_key

        self.config_data["tp_device_parameter"] = sanitize_param_struct(
            "tp_device_parameter", tp_device_parameter, self.config_data, self.config_dir
        )
        self.config_data["tp_target_parameter"] = sanitize_param_struct(
            "tp_target_parameter", tp_target_parameter, self.config_data, self.config_dir
        )

        self._validate()

        if "audit_log" not in self.config_data:
            return

        try:
            self.config_data["audit_log"] = find_file(
                self.config_data["audit_log"],
                search_paths=[self.config_dir] if self.config_dir else None,
            )
        except SPSDKError:
            # file doesn't exist yet
            # if config file is defined make the audit_log path relative to it
            if self.config_dir and "audit_log" in self.SCHEMA_MEMBERS:
                self.config_data["audit_log"] = os.path.join(
                    self.config_dir, self.config_data["audit_log"]
                )

    @property
    def tp_target(self) -> str:
        """Trust Provisioning target."""
        return self.config_data["tp_target"]

    @property
    def tp_target_parameter(self) -> Dict[str, Any]:
        """Trust Provisioning target parameters."""
        return self.config_data["tp_target_parameter"]

    @property
    def audit_log(self) -> str:
        """Path to audit log."""
        return self.config_data["audit_log"]

    @property
    def audit_log_key(self) -> str:
        """Path to audit log key."""
        # up until this point 'audit_log_key' was optional
        if "audit_log_key" not in self.config_data:
            raise SPSDKTpError("audit-log-key is not specified")
        return find_file(
            self.config_data["audit_log_key"],
            search_paths=[self.config_dir] if self.config_dir else None,
        )

    @property
    def firmware_data(self) -> Optional[bytes]:
        """OEM firmware content."""
        if not self.config_data.get("firmware"):
            return None
        file_path = find_file(
            self.config_data["firmware"],
            search_paths=[self.config_dir] if self.config_dir else None,
        )
        return load_binary(file_path)

    @property
    def prov_firmware_data(self) -> Optional[bytes]:
        """Provisioning Firmware content."""
        if not self.config_data.get("prov_firmware"):
            return None
        file_path = find_file(
            self.config_data["prov_firmware"],
            search_paths=[self.config_dir] if self.config_dir else None,
        )
        return load_binary(file_path)


class TPConfigConfig(TPBaseConfig):
    """Configuration class for TPConfig app."""

    SCHEMA_MEMBERS = [
        "family",
        "tp_timeout",
        "device",
        "cmpa",
        "cfpa",
        "sb_kek",
        "user_kek",
        "production_quota",
        "oem_log_prk",
        "nxp_prod_cert",
        "nxp_global_attest_cert",
        "oem_id",
    ]

    def __init__(
        self,
        config_file_path: str,
        tp_device: Optional[str] = None,
        tp_device_parameter: Optional[List[str]] = None,
        timeout: Optional[int] = None,
    ) -> None:
        """Initialize the TPConfig configuration."""
        config_dir = os.path.dirname(config_file_path)
        config_data = self._get_config_data(config_file_path)

        super().__init__(config_data=config_data, config_dir=config_dir)

        if timeout is not None:
            self.config_data["timeout"] = timeout
        if tp_device:
            self.config_data["tp_device"] = tp_device
        self.config_data["tp_device_parameter"] = sanitize_param_struct(
            "tp_device_parameter", tp_device_parameter, self.config_data, self.config_dir
        )

        self._validate()


def multiple_tp_dict(multi: Optional[List[str]]) -> Dict[str, str]:
    """Convert even Multiple option to dict in order.

    :param multi: Input List with multiple options.
    :return: Result dictionary
    :raises SPSDKTpError: Problem parsing the input
    """
    try:
        if multi is None:
            return {}
        return dict(item.split("=") for item in multi if isinstance(item, str))
    except Exception as e:
        raise SPSDKTpError(f"Unable to process input {multi}") from e


def sanitize_param_struct(
    param_name: str, cli_params: Optional[List[str]], config_params: dict, config_dir: Optional[str]
) -> dict:
    """Sanitize TP Target/Device parameter settings.

    :param param_name: Name of the parameter group
    :param cli_params: Parameters passed from CLI
    :param config_params: Parameters read from config file
    :param config_dir: Path to config file (for extended search)
    :return: Sanitized parameter config structure
    """
    sanitized_config = config_params.get(param_name, {}) or {}
    sanitized_config.update(multiple_tp_dict(cli_params))
    if "config_file" in sanitized_config:
        config_file_path = sanitized_config["config_file"]
        config_file_path = find_file(
            config_file_path, search_paths=[config_dir] if config_dir else None
        )
        sanitized_config["config_file"] = config_file_path
    return sanitized_config


def print_device_table(intfs: List[TpIntfDescription]) -> str:
    """Prints the List of Interfaces to nice colored table."""
    if len(intfs) == 0:
        return (
            colorama.Fore.RED
            + "Nothing to print - empty interface list!"
            + colorama.Style.RESET_ALL
        )

    # exhaust the chain iterator
    header = list(
        # get all the header elements from interfaces
        # serialize them in to chain: list[list[str]] -> iterator[str]
        itertools.chain.from_iterable(i.as_dict().keys() for i in intfs),
    )
    # remove duplicates whilst preserving the order
    header = list(dict.fromkeys(header))

    table = prettytable.PrettyTable(_sanitize_table_header(header))
    table.align = "l"
    table.header = True
    table.border = True
    table.hrules = prettytable.HEADER
    table.vrules = prettytable.NONE
    for i, intf in enumerate(intfs):
        fields = [
            colorama.Fore.YELLOW + str(i),
            colorama.Fore.MAGENTA + intf.name,
            colorama.Fore.GREEN + intf.description,
        ]

        for field in header[2:]:
            fields.append(colorama.Fore.CYAN + str(intf.as_dict().get(field, "")))  # type: ignore
        table.add_row(fields)

    return table.get_string() + colorama.Style.RESET_ALL


def _sanitize_table_header(header: List[str]) -> List[str]:
    """Sanitize header for use as Table header.

    Capitalize header items.
    Replace "_" with a " " in header items.
    Insert "#" at the first place.
    """
    assert "name" in header
    assert "description" in header
    header = [item.capitalize() for item in header]
    header = [item.replace("_", " ") for item in header]
    header.insert(0, "#")
    return header


def process_tp_inputs(
    tp_type: str,
    tp_parameters: Union[List[str], Dict],
    scan_func: Callable[[Optional[str], Optional[dict]], List[TpIntfDescription]],
    header: Literal["device", "target"],
    print_func: Callable[[str], None],
) -> TpIntfDescription:
    """Process input from config file and/or command line and return corresponding interface.

    In case no interface (device/target) is found, function throws an error.
    In case there are more interfaces found, print out a list of interfaces and exit.

    :param tp_type: Type of tp adapter (scard, blhost, ...)
    :param tp_parameters: Parameters specifying TP Interface
    :param scan_func: Function for scanning TP Interfaces
    :param header: String determining potential error message header (device/target)
    :param print_func: Function for displaying potential error message
    :raises SPSDKTpError: No TP Interface (device/target) found
    :return: Designated Interface Description
    """
    assert tp_type in get_tp_device_types() + get_tp_target_types()
    header_map = {"device": "TP Device", "target": "TP Target"}
    styled_header = header_map[header]
    params = tp_parameters if isinstance(tp_parameters, dict) else multiple_tp_dict(tp_parameters)
    interfaces = scan_func(tp_type, params)
    if not interfaces:
        raise SPSDKTpError(
            f"No {styled_header} found. Search criteria: type={tp_type}; params={params}"
        )
    if len(interfaces) > 1:
        print_func(
            f"{len(interfaces)} {styled_header}s found "
            f"with search criteria type={tp_type}; params={params}"
        )
        print_func(print_device_table(interfaces))
        print_func(f"You need to provide search criteria that fit only one {styled_header}.")
        raise SPSDKTpError()
    return interfaces[0]


###############################################################
# Common options/commands used in both tphost and tpconfig applications


def tp_device_options(options: FC) -> FC:
    """Provides: tp_device: str, tp_device_parameter: List[str]."""
    options = click.option(
        "-dp",
        "--tp-device-parameter",
        type=str,
        multiple=True,
        help="""
The Trusted Device parameters, may be used multiple times to setup TP device interface.
To see available parameters for given TP device, run `tphost/tpconfig device-help`.
Example: -dp id=123456
""",
    )(options)
    options = click.option(
        "-d",
        "--tp-device",
        type=click.Choice(get_tp_device_types(), case_sensitive=False),
        help="The Trusted Device to be used for provisioning target.",
        default=get_tp_device_types()[0] if single_tp_device_adapter() else None,
    )(options)
    return options


def tp_target_options(options: FC) -> FC:
    """Provides: tp_target: str, tp_target_parameter: List[str]."""
    options = click.option(
        "-tp",
        "--tp-target-parameter",
        type=str,
        multiple=True,
        help="""
The Trusted Target parameter, may be used multiple times to setup TP target interface.
To see available parameters for given TP device, run `tphost/tpconfig target-help`.
Example: -tp blhost_baudrate=115200
""",
    )(options)
    options = click.option(
        "-t",
        "--tp-target",
        type=click.Choice(get_tp_target_types(), case_sensitive=False),
        help="The Trusted Target Interface to be used for provisioning process.",
        default=get_tp_target_types()[0] if single_tp_target_adapter() else None,
    )(options)
    return options


@click.command(name="device-help")
@click.option(
    "-d",
    "--tp-device",
    help="Device name to print help, if not used, all devices help will be printed.",
)
def device_help(tp_device: Optional[str] = None) -> None:
    """Command prints help for all devices or optionally only for specified."""
    dev_list = [tp_device] if tp_device else get_tp_device_types()

    for i, dev in enumerate(dev_list):
        click.echo(colorama.Fore.YELLOW + f"#{i}: {dev.upper()} device:")
        click.echo(colorama.Fore.WHITE + get_tp_device_class(dev).get_help() + "\n")
    click.echo(colorama.Style.RESET_ALL)


@click.command(name="target-help")
@click.option(
    "-t",
    "--tp-target",
    help="Target name to print help, if not used, all targets help will be printed.",
)
def target_help(tp_target: Optional[str] = None) -> None:
    """Command prints help for all targets or optionally only for specified."""
    target_list = [tp_target] if tp_target else get_tp_target_types()

    for i, target in enumerate(target_list):
        click.echo(colorama.Fore.YELLOW + f"#{i}: {target.upper()} target:")
        click.echo(colorama.Fore.WHITE + get_tp_target_class(target).get_help() + "\n")
    click.echo(colorama.Style.RESET_ALL)


@click.command(name="list-tptargets", no_args_is_help=not single_tp_target_adapter())
@tp_target_options
def list_tptargets(tp_target: str, tp_target_parameter: List[str]) -> None:
    """Command prints all supported and connected TP targets."""
    tp_targets = scan_tp_targets(tp_target, multiple_tp_dict(tp_target_parameter))
    click.echo(print_device_table(tp_targets))


@click.command(name="list-tpdevices", no_args_is_help=not single_tp_device_adapter())
@tp_device_options
def list_tpdevices(tp_device: str, tp_device_parameter: List[str]) -> None:
    """Command prints all supported and connected TP devices."""
    tp_devices = scan_tp_devices(tp_device, multiple_tp_dict(tp_device_parameter))
    click.echo(print_device_table(tp_devices))


@click.command(name="get-counters", no_args_is_help=False)
@tp_device_options
@click.option(
    "-t",
    "--timeout",
    type=click.IntRange(0, 600, clamp=True),
    help="The target provisioning timeout in seconds.",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to configuration file (parameters on CLI take precedence).",
    required=False,
)
def get_counters(
    tp_device: str,
    tp_device_parameter: List[str],
    timeout: int,
    config: str,
) -> None:
    """Get Provisioning counters from TP device."""
    TPHostConfig.SCHEMA_MEMBERS = ["device"]
    tp_config = TPHostConfig(
        tp_device=tp_device,
        tp_device_parameter=tp_device_parameter,
        config=config,
        timeout=timeout,
    )
    device = tp_config.tp_device
    params = tp_config.tp_device_parameter

    tp_interface = process_tp_inputs(
        tp_type=device,
        tp_parameters=params,
        header="device",
        scan_func=scan_tp_devices,
        print_func=click.echo,
    )
    tp_dev = tp_interface.create_interface()
    assert isinstance(tp_dev, TpDevInterface)

    tp_worker = TrustProvisioningHost(
        tpdev=tp_dev,
        tptarget=None,  # type: ignore  # target is not used, we set it to None on purpose
        info_print=click.echo,
    )
    tp_worker.get_counters()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for cryptographic utilities."""

import os

from spsdk.utils import UTILS_DATA_FOLDER

CRYPTO_SCH_FILE: str = os.path.join(UTILS_DATA_FOLDER, "sch_crypto.yaml")
OTFAD_DATA_FOLDER: str = os.path.join(UTILS_DATA_FOLDER, "otfad")
OTFAD_SCH_FILE: str = os.path.join(OTFAD_DATA_FOLDER, "sch_otfad.yaml")
OTFAD_DATABASE_FILE: str = os.path.join(OTFAD_DATA_FOLDER, "database.yaml")

IEE_DATA_FOLDER: str = os.path.join(UTILS_DATA_FOLDER, "iee")
IEE_SCH_FILE: str = os.path.join(IEE_DATA_FOLDER, "sch_iee.yaml")
IEE_DATABASE_FILE: str = os.path.join(IEE_DATA_FOLDER, "database.yaml")


from .abstract import BackendClass
from .cert_blocks import CertBlock, CertBlockV2, CertBlockV31
from .certificate import Certificate
from .common import Counter, crypto_backend, matches_key_and_cert, serialize_ecc_signature
from .iee import (
    Iee,
    IeeKeyBlob,
    IeeKeyBlobAttribute,
    IeeKeyBlobKeyAttributes,
    IeeKeyBlobLockAttributes,
    IeeKeyBlobModeAttributes,
    IeeKeyBlobWritePmsnAttributes,
)
from .otfad import KeyBlob, Otfad

__all__ = [
    "matches_key_and_cert",
    "crypto_backend",
    "Counter",
    "BackendClass",
    "CertBlockV2",
    "Certificate",
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides exceptions for PFR module."""
from spsdk import SPSDKError


class SPSDKPfrError(SPSDKError):
    """General PFR error."""


class SPSDKPfrConfigError(SPSDKPfrError):
    """General PFR configuration error."""


class SPSDKPfrConfigReadError(SPSDKPfrConfigError):
    """Configuration file decode error."""


class SPSDKPfrRotkhIsNotPresent(SPSDKPfrError):
    """The configuration area doesn't provide ROTKH field."""


class SPSDKPfrcMissingConfigError(SPSDKPfrError):
    """The translation of conditions failed."""



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Commands for Debug Mailbox."""
import time
from typing import Any, List, Optional

from spsdk import SPSDKError
from spsdk.utils.exceptions import SPSDKTimeoutError
from spsdk.utils.misc import format_value

from .debug_mailbox import DebugMailbox, logger


class DebugMailboxCommand:
    """Class for DebugMailboxCommand."""

    STATUS_IS_DATA_MASK = 0x00
    # default delay after sending a command, in seconds
    DELAY_DEFAULT = 0.03

    def __init__(
        self,
        dm: DebugMailbox,
        id: int,  # pylint: disable=redefined-builtin
        name: str = "",
        paramlen: int = 0,
        resplen: int = 0,
        delay: float = DELAY_DEFAULT,
    ):
        """Initialize."""
        self.dm = dm
        self.paramlen = paramlen
        self.resplen = resplen
        self.id = id
        self.name = name
        self.delay = delay

    def run(self, params: Optional[List[int]] = None) -> List[Any]:
        """Run DebugMailboxCommand."""
        paramslen = len(params) if params else 0
        if paramslen != self.paramlen:
            raise SPSDKError(
                "Provided parameters length is not equal to command parameters length!"
            )

        req = self.id | (self.paramlen << 16)
        logger.debug(f"<- spin_write: {format_value(req, 32)}")
        self.dm.spin_write(self.dm.registers["REQUEST"]["address"], req)

        # Wait to allow reset of internal logic of debug mailbox
        time.sleep(self.delay)

        if params:
            for i in range(paramslen):
                ret = self.dm.spin_read(self.dm.registers["RETURN"]["address"])
                logger.debug(f"-> spin_read:  {format_value(ret, 32)}")
                if (ret & 0xFFFF) != 0xA5A5:
                    raise SPSDKError("Device did not send correct ACK answer!")
                if ((ret >> 16) & 0xFFFF) != (self.paramlen - i):
                    raise SPSDKError(
                        "Device expects parameters of different length we can provide!"
                    )
                logger.debug(f"<- spin_write: {format_value(params[i], 32)}")
                self.dm.spin_write(self.dm.registers["REQUEST"]["address"], params[i])

        ret = self.dm.spin_read(self.dm.registers["RETURN"]["address"])
        logger.debug(f"-> spin_read:  {format_value(ret, 32)}")

        # bit 31 is flag in new protocol version
        # new_protocol = bool(ret >> 31)
        error_indication = bool(ret >> 20) and not bool(self.resplen)
        # solve the case that response is in legacy protocol and there is some
        # unwanted bits in none expected data. In this case return valid read data.
        if not self.resplen and not error_indication:
            return [ret]

        resplen = (ret >> 16) & 0x7FFF
        status = ret & 0xFFFF

        if status != 0:
            raise SPSDKError(f"Status code is not success: {status} !")

        if resplen != self.resplen:  # MSB is used to show it is the new protocol -> 0x7FFF
            raise SPSDKError("Device wants to send us different size than expected!")

        # do not send ack, in case no data follows
        if resplen == 0:
            return []

        # ack the response
        ack = 0xA5A5 | (self.resplen << 16)
        logger.debug(f"<- spin_write: {format_value(ack, 32)}")
        self.dm.spin_write(self.dm.registers["REQUEST"]["address"], ack)

        response = []
        for i in range(self.resplen):
            ret = self.dm.spin_read(self.dm.registers["RETURN"]["address"])
            logger.debug(f"-> spin_read:  {format_value(ret, 32)}")
            response.append(ret)
            ack = 0xA5A5 | ((self.resplen - i - 1) << 16)
            logger.debug(f"<- spin_write: {format_value(ack, 32)}")
            self.dm.spin_write(self.dm.registers["REQUEST"]["address"], ack)
        return response

    def run_safe(self, raise_if_failure: bool = True, **args: Any) -> Optional[List[Any]]:
        """Run a command and abort on first failure instead of looping forever."""
        try:
            return self.run(**args)
        except (SPSDKTimeoutError, TimeoutError) as error:
            if raise_if_failure:
                raise SPSDKTimeoutError("Timeout occurred") from error
            logger.error(str(error))
        return None


class StartDebugMailbox(DebugMailboxCommand):
    """Class for StartDebugMailbox."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=1, name="START_DBG_MB")


class GetCRPLevel(DebugMailboxCommand):
    """Class for Get CRP Level."""

    # Set STATUS_IS_DATA_MASK to range 0-255, because larger life cycle is not expected
    STATUS_IS_DATA_MASK = 0xFF

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=2, name="GET_CRP_LEVEL")


class EraseFlash(DebugMailboxCommand):
    """Class for Erase Flash."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=3, name="ERASE_FLASH", delay=0.5)


class ExitDebugMailbox(DebugMailboxCommand):
    """Class for ExitDebugMailbox."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=4, name="EXIT_DBG_MB")


class EnterISPMode(DebugMailboxCommand):
    """Class for EnterISPMode."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=5, name="ENTER_ISP_MODE", paramlen=1)


class EnterBlankDebugAuthentication(DebugMailboxCommand):
    """Class for EnterBlankDebugAuthentication."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=0x8, name="ENTER_BLANK_DEBUG_AUTH", paramlen=8)


class SetFaultAnalysisMode(DebugMailboxCommand):
    """Class for SetFaultAnalysisMode."""

    def __init__(self, dm: DebugMailbox, paramlen: int = 0) -> None:
        """Initialize."""
        super().__init__(dm, id=6, name="SET_FA_MODE", paramlen=paramlen)


class StartDebugSession(DebugMailboxCommand):
    """Class for StartDebugSession."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=7, name="START_DBG_SESSION")


class DebugAuthenticationStart(DebugMailboxCommand):
    """Class for DebugAuthenticationStart."""

    def __init__(self, dm: DebugMailbox, resplen: int = 26) -> None:
        """Initialize."""
        # 26 words == 104 bytes (SHA256 - 32 Bytes)
        # 30 words == 120 bytes (SHA384 - 48 Bytes)
        super().__init__(dm, id=16, name="DBG_AUTH_START", resplen=resplen)


class DebugAuthenticationResponse(DebugMailboxCommand):
    """Class for DebugAuthenticationResponse."""

    def __init__(self, dm: DebugMailbox, paramlen: int) -> None:
        """Initialize."""
        super().__init__(dm, id=17, name="DBG_AUTH_RESP", paramlen=paramlen)


class NxpDebugAuthenticationStart(DebugMailboxCommand):
    """Class for DebugAuthenticationStart."""

    def __init__(self, dm: DebugMailbox, resplen: int = 26) -> None:
        """Initialize."""
        # 26 words == 104 bytes (SHA256 - 32 Bytes)
        # 30 words == 120 bytes (SHA384 - 48 Bytes)
        super().__init__(dm, id=18, name="NXP_DBG_AUTH_START", resplen=resplen)


class NxpDebugAuthenticationResponse(DebugMailboxCommand):
    """Class for DebugAuthenticationResponse."""

    def __init__(self, dm: DebugMailbox, paramlen: int) -> None:
        """Initialize."""
        super().__init__(dm, id=19, name="NXP_DBG_AUTH_RESP", paramlen=paramlen)


class StartDebugSessions(DebugMailboxCommand):
    """Class for StartDebugSessions."""

    def __init__(self, dm: DebugMailbox) -> None:
        """Initialize."""
        super().__init__(dm, id=7, name="START_DEBUG_SESSION")
